;; guile-td.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(use-modules (srfi srfi-1)
             (aa-tree)
             (oop goops)
             (chickadee)
             (chickadee math vector)
             (chickadee render sprite)
             (chickadee render texture)
             (system repl server)
             (system repl coop-server))

(read-set! keywords 'prefix)

(define-class <euclidean> ()
  (x :init-value 10 :getter get-x :init-keyword :x)
  (y :init-value 10 :getter get-y :init-keyword :y))

(define-class <drawable> (<euclidean>)
  (sprite-key-proc :getter get-sprite-key-proc :init-keyword :sprite-key-proc))

(define-method (get-sprite-key (d <drawable>))
  ((get-sprite-key-proc d) d))

(define-class <marching> (<drawable>)
  (facing :init-value :left :getter get-facing :init-keyword :facing))

(define world #nil)

(define (load-field)
  (set! world
    (aa-insert world '(:sprites :field) (load-image "images/field01.png"))))

(define (load-units)
  (set! world
    (fold
     (lambda (lst w)
       (aa-insert w (first lst) (primitive-eval (second lst)))) ;; BUG, need to eval load-image
     world
     '(
       ((:sprites :infantry-left) (load-image "images/black-arrow-left.png"))
       ((:sprites :infantry-right) (load-image "images/black-arrow-right.png"))
       ((:sprites :infantry-up) (load-image "images/black-arrow-up.png"))
       ((:sprites :infantry-down) (load-image "images/black-arrow-down.png"))
       ))))

(define (load-sprites)
  (load-field)
  (load-units)
  )

(define (nil-guard val thunk proc)
  (if (nil? val)
      (thunk)
      (proc val)))

(define (nil-thunk) #nil)

(define (id x) x)

(define (draw-sprites alpha)
  (draw-field alpha)
  (let ((enemy-list
         (nil-guard (aa-search world '(:enemies))
                    nil-thunk
                    (lambda (r) (car r)))))
    (for-each
     (lambda (drawable)
       (let ((sprite (car (aa-search world
                                     (list
                                      :sprites (get-sprite-key drawable))))))
         (draw-sprite
          sprite
          (vec2 (get-x drawable) (get-y drawable)))))
     enemy-list)))
        
(define (draw-field alpha)
  (draw-sprite
   (car (aa-search world '(:sprites :field)))
   (vec2 0.0 0.0)))

(define generic-handler
  (lambda (key . args)
    (simple-format #t "exception key: ~S\n" key)
    (display "keys: ")
    (for-each
     (lambda (a) (simple-format #t "~S\n" a)) args)))

(define generic-unwind
  (lambda (key . args)
    (backtrace)))

(define (generic-catch proc)
         (catch #t proc generic-handler generic-unwind))

(define (start-td)
  (let* ((server-socket (make-tcp-server-socket))
          (coop-server (spawn-coop-repl-server server-socket)))
    (add-hook! update-hook
               (lambda (dt)
                 (poll-coop-repl-server coop-server)))
    (add-hook! quit-hook (lambda () (abort-game)))
    (add-hook! load-hook (lambda () (load-sprites)))
    (add-hook! draw-hook
               (lambda (alpha)
                 (generic-catch
                   (lambda ()
                     (draw-sprites alpha)))))
    (run-game :window-title "Guile TD"
              :update-hz 10)
    (stop-server-and-clients!)))
